// ==============================================================
// MFDlib.cpp
//
// Routines to assist in the development of text based MFD's
// for Orbiter
//
// (c) 2002 Trevor Johns
// trevorjohns@yahoo.com
// ==============================================================

#define STRICT
#include <windows.h>
#include <stdio.h>
#include <stdarg.h>
#include "mfdlib.h"

// MFDFontSize : size (pixels) of MFD fixed pitch font
SIZE MFDFontSize;	

// MFDUpdate(hDC, cw, ch)
// initialize MFD details for text output
void MFDUpdate(const HDC hDC, const int cw, const int ch)
{
	MFDFontSize.cx = cw;
	MFDFontSize.cy = ch;
	SetTextAlign(hDC, TA_TOP | TA_LEFT | TA_UPDATECP);
	SetTextXY(hDC, 0, 0);
}

// p = GetTextPos(hDC)
// returns the current (pixel) position for text output
POINT GetTextPos(const HDC hDC)
{
POINT p;

	MoveToEx(hDC, 0, 0, &p);		// move to (0,0) & get current position
	MoveToEx(hDC, p.x, p.y, NULL);	// restore position
	return p;
}

// p = GetTextXY(hDC)
// returns the current (char) position for text output
POINT GetTextXY(const HDC hDC)
{
POINT p;

	// get current (pixel) position
	p = GetTextPos(hDC);
	// convert current position from pixel => char
	// NB: text is offset 1/2 char for left margin
	p.x = (p.x - MFDFontSize.cx/2) / MFDFontSize.cx;
	p.y = p.y / MFDFontSize.cy;
	return p;
}

// oldp = SetTextPos(hDC, x, y)
// move the current (pixel) position for text output
// returns old (pixel) position
POINT SetTextPos(const HDC hDC, const int x, const int y)
{
POINT oldp;

	MoveToEx(hDC, x, y, &oldp);
	return oldp;
}

// oldp = SetTextXY(hDC, x, y)
// move the current (char) position for text output
// returns old (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
POINT SetTextXY(const HDC hDC, const int x, const int y)
{
POINT oldp, p;

	// NB: text is offset 1/2 char for left margin
	oldp = p = GetTextPos(hDC);
	if (x >= 0) p.x = x*MFDFontSize.cx+MFDFontSize.cx/2;
	if (y >= 0) p.y = y*MFDFontSize.cy;
	MoveToEx(hDC, p.x, p.y, NULL);
	// convert old position from pixel => char
	oldp.x = (oldp.x - MFDFontSize.cx/2) / MFDFontSize.cx;
	oldp.y = oldp.y / MFDFontSize.cy;
	return oldp;
}

// set the MFD foreground color for future text
// NB: background becomes transparent
// returns previous foreground color
COLORREF SetMFDColor(const HDC hDC, const COLORREF fg)
{
	SetBkMode(hDC, TRANSPARENT);
	return SetTextColor(hDC, fg);
}

// set the MFD foreground & background color for future text
// returns previous foreground color
COLORREF SetMFDColor(const HDC hDC, const COLORREF fg, const COLORREF bg)
{
	SetBkMode(hDC, OPAQUE);	
	SetBkColor(hDC, bg);
	return SetTextColor(hDC, fg);
}

// ===========================================
// text printing routine come in 3 flavours
// (X, Y) position for printing
// foreground color + (X, Y) position
// foreground + background + (X, Y) position
// ===========================================

// oldp = TextXY(hDC, x, y, fmt [, ...]);
// print a formatted string (with parameters) at a specified (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const int x, const int y, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// move to output position & print buffer
	p = SetTextXY(hDC, x, y);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}

// oldp = TextXYC(hDC, x, y, fg, fmt [, ...]);
// print a formatted color string (with parameters) at a specified (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const int x, const int y, const COLORREF fg, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// move to output position, set color & print buffer
	p = SetTextXY(hDC, x, y);
	SetMFDColor(hDC, fg);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}

// oldp = TextXYC(hDC, x, y, fg, bg, fmt [, ...]);
// print a formatted color string (with parameters) at a specified (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const int x, const int y, const COLORREF fg, const COLORREF bg, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// move to output position, set color & print buffer
	p = SetTextXY(hDC, x, y);
	SetMFDColor(hDC, fg, bg);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}

// ===========================================
// text printing routine come in 3 MORE flavours
// use current position for printing
// foreground color + current position
// foreground + background + current position
// ===========================================

// oldp = TextXY(hDC, fmt [, ...]);
// print a formatted string (with parameters) at current (char) position
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// get current position & print buffer
	p = GetTextXY(hDC);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}

// oldp = TextXYC(hDC, fg, fmt [, ...]);
// print a formatted color string (with parameters) at current (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const COLORREF fg, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// get current position, set color & print buffer
	p = GetTextXY(hDC);
	SetMFDColor(hDC, fg);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}

// oldp = TextXYC(hDC, fg, bg, fmt [, ...]);
// print a formatted color string (with parameters) at current (char) position
// if x = USE_CURRENT, x position is unchanged (current column)
// if y = USE_CURRENT, y position is unchanged (current row)
// returns old (char) position
CDECL POINT TextXY(const HDC hDC, const COLORREF fg, const COLORREF bg, const char *s, ...)
{
char buf[250];
va_list arg_ptr;
POINT p;

	// create output buffer from format string + arguments
	va_start(arg_ptr, s);
	vsprintf(buf, s, arg_ptr);
	va_end(arg_ptr);

	// get current position, set color & print buffer
	p = GetTextXY(hDC);
	SetMFDColor(hDC, fg, bg);
	TextOut(hDC, 0, 0, buf, lstrlen(buf));
	return p;
}
