#include "FlightStatus.h"

const double FT = 3.2808399;
const double KT = 1.94384449;

FlightStatus::FlightStatus (VESSEL *vessel) 
: vessel (vessel) {
	heading = 0;
	prevKIAS = prevKIASt = KIASacc = 0.0;
	gFactor = 1.0;
	prevRvel.x = prevRvel.y = prevRvel.z = 0.0;
	prevSimt = 0.0;
}

void FlightStatus::TimeStep (const double simt) {
	VECTOR3 v;
	altitude = vessel->GetAltitude();
	vessel->GetShipAirspeedVector(v);
	forwardvel = max (0, v.z);
	airspeed = vessel->GetAirspeed();
	pitch = vessel->GetPitch()*DEG;
	bank = vessel->GetBank()*DEG;
	oapiGetHeading(vessel->GetHandle(), &heading);
	heading *= DEG;
	vessel->GetAngularVel(v);
	pitchrate = v.x*DEG; rollrate = v.z*DEG; yawrate = -v.y*DEG;
	mach = vessel->GetMachNumber();
	alpha = vessel->GetAOA()*DEG;
	beta = vessel->GetSlipAngle()*DEG;
	vessel->GetHorizonAirspeedVector(horizonVel);
	vertvel = horizonVel.y;
	vvFt = vertvel * 60 * FT;
	throttle = vessel->GetEngineLevel(ENGINE_MAIN);

	altFt = altitude * FT;

	double cs;
	if (vessel->GetAtmDensity() < 1e-6)
		cs = 0.0;
	else
		cs = KT * sqrt ((7.0 / 5.0) * vessel->GetAtmPressure() / vessel->GetAtmDensity());

	if (cs > 1e-6)
		KIAS = KT * forwardvel * (661.47 * sqrt (vessel->GetAtmPressure() / ATMP ) / cs );
	else
		KIAS = 0.0;

	if (simt - prevKIASt > 0.1) {
		KIASacc = (KIAS - prevKIAS) / (simt - prevKIASt);
		prevKIAS = KIAS;
		prevKIASt = simt;
	}

	gFactor = CalculateGFactor(simt);
}

double FlightStatus::CalculateGFactor(const double simt) {
	VESSELSTATUS vs;

	vessel->GetStatus(vs);

	double delta_t = simt - prevSimt;

	VECTOR3 acc;
	acc.x = (vs.rvel.x - prevRvel.x) / delta_t;
	acc.y = (vs.rvel.y - prevRvel.y) / delta_t;
	acc.z = (vs.rvel.z - prevRvel.z) / delta_t;

	VECTOR3 gf = _V(vs.rpos.x,vs.rpos.y,vs.rpos.z);
	double mod = sqrt (gf.x*gf.x+gf.y*gf.y+gf.z*gf.z);
	double r = mod / 1000;
	gf.x /= mod; gf.y /= mod; gf.z /= mod;

	gf *= 397779730 / r / r; //this is g vector acceleration in m/sec
	gf = acc + gf; //the diference between the two is actual G press
	
	prevRvel = vs.rvel;
	prevSimt = simt;

	return sqrt (gf.x*gf.x+gf.y*gf.y+gf.z*gf.z) / G;
}