// ==============================================================
//                 ORBITER MODULE: DialogTemplate
//                  Part of the ORBITER SDK
//            Copyright (C) 2003 Martin Schweiger
//                   All rights reserved
//
// SpaceShipOneMFD.cpp
//
// This module demonstrates how to build an Orbiter plugin which
// inserts a new MFD (multi-functional display) mode. The code
// is not very useful in itself, but it can be used as a starting
// point for your own MFD developments.
// ==============================================================

#define STRICT
#define ORBITER_MODULE
#include "windows.h"
#include "orbitersdk.h"
#include "SpaceShipOneMFD.h"
#include "MFDlib.h"
#include "utils/FloatUtils.h"

// ==============================================================
// Global variables

static struct {  
	int mode; 
    SpaceShipOneMFD *CurrentMFD;
} g_SpaceShipOneMFD;

static struct {
	HPEN   pScale;
	HBRUSH bScale;
	HPEN   pPointer;
	HBRUSH bPointer;
	HPEN   pArcPos;
	HPEN   pArcVPos;
	HPEN   pArcNeg;
	HPEN   pArcVNeg;
	HPEN   pWing;
	HPEN   pFP;
	HPEN   pPred;
} g_GDI;
int g_MFDmode; // identifier for new MFD mode

// ==============================================================
// API interface

DLLCLBK void opcDLLInit (HINSTANCE hDLL) {
	static char *name = "SpaceShipOneMFD";   // MFD mode name
	MFDMODESPEC spec;
	spec.name = name;
	spec.key = OAPI_KEY_Z;                // MFD mode selection key
	spec.msgproc = SpaceShipOneMFD::MsgProc;  // MFD mode callback function

	// Register the new MFD mode with Orbiter
	g_SpaceShipOneMFD.mode = oapiRegisterMFDMode (spec);
	g_SpaceShipOneMFD.CurrentMFD = NULL;

	g_GDI.pScale	= CreatePen (PS_SOLID, 1, WHITE);
	g_GDI.bScale	= CreateSolidBrush (LIGHTWHITE);
	g_GDI.pPointer	= CreatePen (PS_SOLID, 1, GREEN);
	g_GDI.bPointer	= CreateSolidBrush (LIGHTGREEN);
	g_GDI.pArcPos	= CreatePen (PS_SOLID | PS_ENDCAP_ROUND, 3, LIGHTGREEN);
	g_GDI.pArcNeg	= CreatePen (PS_SOLID | PS_ENDCAP_ROUND, 3, LIGHTRED);
	g_GDI.pWing		= CreatePen (PS_SOLID | PS_ENDCAP_ROUND, 2, BRIGHTWHITE);
	g_GDI.pFP		= CreatePen (PS_SOLID | PS_ENDCAP_ROUND, 2, BRIGHTGREEN);
	g_GDI.pPred		= CreatePen (PS_SOLID | PS_ENDCAP_ROUND, 2, GREEN);
}

DLLCLBK void opcDLLExit (HINSTANCE hDLL) {
	// Unregister the custom MFD mode when the module is unloaded
	oapiUnregisterMFDMode (g_SpaceShipOneMFD.mode);
	DeleteObject (g_GDI.pScale);
	DeleteObject (g_GDI.bScale);
	DeleteObject (g_GDI.pPointer);
	DeleteObject (g_GDI.bPointer);
	DeleteObject (g_GDI.pArcPos);
	DeleteObject (g_GDI.pArcNeg);
	DeleteObject (g_GDI.pWing);
	DeleteObject (g_GDI.pFP);
	DeleteObject (g_GDI.pPred);
}

DLLCLBK void opcTimestep(double SimT, double SimDT, double mjd) {
	// Don't try to call the function before the class has been created! :-)
	if (g_SpaceShipOneMFD.CurrentMFD != NULL) { 
		g_SpaceShipOneMFD.CurrentMFD->TimeStep(SimT);
	}
}

// ==============================================================
// MFD class implementation

// MFD message parser
int SpaceShipOneMFD::MsgProc (UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam) {
	switch (msg) {
	case OAPI_MSG_MFD_OPENED:
		// Our new MFD mode has been selected, so we create the MFD and
		// return a pointer to it.
		return (int)(new SpaceShipOneMFD (LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
	}
	return 0;
}

// Constructor
SpaceShipOneMFD::SpaceShipOneMFD (DWORD w, DWORD h, VESSEL *vessel)
: MFD (w, h, vessel) {
	g_SpaceShipOneMFD.CurrentMFD = this;
	fs = new FlightStatus (pV);
	adi = new ADI (1, 1, w-2, h*2/3);
	zoom = 2;
}

SpaceShipOneMFD::~SpaceShipOneMFD () {
	delete (fs);
	delete (adi);
	//Added by JRB 2/11/07
	g_SpaceShipOneMFD.CurrentMFD = NULL;
}

void SpaceShipOneMFD::TimeStep (double simT) {
	fs->TimeStep (simT);
}

bool SpaceShipOneMFD::ConsumeKeyBuffered(DWORD key) {
	switch(key) {
	case OAPI_KEY_V:
		zoom++;
		if (zoom > 4) zoom = 1;
		return true;
	default:
		return false;
	}
}

// Repaint the MFD
void SpaceShipOneMFD::Update (HDC hDC) {
	
	MFDUpdate (hDC, cw, ch);

	DrawADI (hDC);
	DrawAirspeed (hDC);
	DrawAltitude (hDC);
	DrawFlightData (hDC);
	DrawApogeeData (hDC);
	DrawPropellantData (hDC);
	DrawRates (hDC);
}

void SpaceShipOneMFD::DrawADI(HDC hDC) {

	adi->Draw (hDC, (float)fs->pitch, (float)-fs->bank, (float)-fs->heading, (float)fs->alpha, (float)fs->beta, (float)fs->pitchrate, (float)fs->yawrate, (float)zoom);

	SetMFDColor (hDC, WHITE);

	if (fs->heading > 267 || fs->heading < 93 || abs(fs->pitch) > 80) {
		SetTextPos (hDC,
			(int)(adi->x + adi->NSEW[0] - cw/2),
			(int)(adi->y + adi->NSEW[1] - ch/2));
		TextXY (hDC, "N");

	}

	if (fs->heading > 87 && fs->heading < 273 || abs(fs->pitch) > 80) {
		SetTextPos (hDC,
			(int)(adi->x + adi->NSEW[2] - cw/2),
			(int)(adi->y + adi->NSEW[3] - ch/2));
		TextXY (hDC, "S");
	}

	if (fs->heading > 357 || fs->heading < 183 || abs(fs->pitch) > 80) {
		SetTextPos (hDC,
			(int)(adi->x + adi->NSEW[4] - cw/2),
			(int)(adi->y + adi->NSEW[5] - ch/2));
		TextXY (hDC, "E");
	}

	if (fs->heading > 177 || fs->heading < 3 || abs(fs->pitch) > 80) {
		SetTextPos (hDC,
			(int)(adi->x + adi->NSEW[6] - cw/2),
			(int)(adi->y + adi->NSEW[7] - ch/2));
		TextXY (hDC, "W");
	}

	SelectObject (hDC, g_GDI.pWing);
	MoveToEx (hDC, W/2 - cw*4, H/3, 0);
	LineTo (hDC, W/2 - cw*3/2, H/3);
	LineTo (hDC, W/2, H/3 + ch);
	LineTo (hDC, W/2 + cw*3/2, H/3);
	LineTo (hDC, W/2 + cw*4, H/3);

	SelectObject (hDC, g_GDI.pFP);
	int x, y;
	x = RoundValue((float)adi->vVector[0]); y = RoundValue((float)adi->vVector[1]);
	Ellipse (hDC, x-cw, y-cw, x+cw, y+cw);
	MoveToEx (hDC, x-cw, y, 0);
	LineTo (hDC, x-cw*2, y);
	MoveToEx (hDC, x+cw, y, 0);
	LineTo (hDC, x+cw*2, y);
	MoveToEx (hDC, x, y-cw, 0);
	LineTo (hDC, x, y-cw*2);

	SelectObject (hDC, g_GDI.pPred);
	MoveToEx (hDC, W/2, H/3, 0);
	x = RoundValue((float)adi->t10Vector[0]); y = RoundValue((float)adi->t10Vector[1]);
	LineTo (hDC, x, y);
	Ellipse (hDC, x-cw/2, y-cw/2, x+cw/2, y+cw/2);

}

void SpaceShipOneMFD::DrawAirspeed(HDC hDC) {
	float KIAS = (float)fs->KIAS;
	float acc  = (float)fs->KIASacc;
	CheckRange (acc, -99.0f, 99.0f);

	int x, y, r;

	x = W/6 + 6;
	y = H/6 + 6;
	r = W/6;

	SetMFDColor (hDC, WHITE);
	SetTextPos (hDC, x - cw*3/2, y - ch/2);
	TextXY (hDC, "%3.0f", fs->KIAS);
	SetMFDColor (hDC, LIGHTWHITE);
	SetTextPos (hDC, x - cw*3/2, y + ch/2);
	TextXY (hDC, "%+2.0f", acc);

	DrawArc (hDC, x, y, r+2, acc * 18, g_GDI.pArcPos, g_GDI.pArcNeg);
	DrawScale (hDC, x, y, r, g_GDI.pScale, g_GDI.bScale);
	DrawPointer (hDC, x, y, r, KIAS * 3.6f, g_GDI.pPointer, g_GDI.bPointer);
}
void SpaceShipOneMFD::DrawAltitude(HDC hDC) {
	float alt = (float)fs->altFt;
	float vv = (float)fs->vvFt;

	int x, y, r;

	x = W*5/6 - 6;
	y = H/6 + 6;
	r = W/6;

	CheckRange (vv, -99999.0f, 99999.0f);

	if (alt > 10000)
		alt = (float)((int)(alt/100) * 100);


	float angle = vv * 0.018f;

	if (abs(vv) > 10000) {
		vv = (float)((int)(vv/100) * 100);
		angle /= 10;
	}

	SetMFDColor (hDC, WHITE);
	SetTextPos (hDC, x - cw*3, y - ch/2);
	TextXY (hDC, "%6.0f", alt);
	SetMFDColor (hDC, LIGHTWHITE);
	SetTextPos (hDC, x - cw*3, y + ch/2);
	TextXY (hDC, "%+5.0f", vv);


	DrawArc (hDC, x, y, r+2, angle, g_GDI.pArcPos, g_GDI.pArcNeg);

	DrawScale (hDC, x, y, r, g_GDI.pScale, g_GDI.bScale);

	angle = alt * 0.36f;

	if (alt > 10000)
		angle /= 10;

	DrawPointer (hDC, x, y, r, angle, g_GDI.pPointer, g_GDI.bPointer);
}

void SpaceShipOneMFD::DrawFlightData(HDC hDC) {
	SetMFDColor (hDC, WHITE);
	SetTextPos (hDC, cw, H*1/3 + ch*1);
	TextXY (hDC, "M %1.2f", fs->mach);
	SetTextPos (hDC, cw, H*1/3 + ch*2);
	TextXY (hDC, "T %3.0f", fs->airspeed*1.94384449);
	SetTextPos (hDC, cw, H*1/3 + ch*3);
	TextXY (hDC, "G %+1.1f", fs->gFactor);
	SetTextPos (hDC, cw, H*1/3 + ch*4);
	TextXY (hDC, "a %+1.1f", fs->alpha);
	SetTextPos (hDC, cw, H*1/3 + ch*5);
	TextXY (hDC, "b %+1.1f", fs->beta);
	SetTextPos (hDC, cw, H*1/3 + ch*6);
	TextXY (hDC, "p %+1.1f", fs->pitchrate);
	SetTextPos (hDC, cw, H*1/3 + ch*7);
	TextXY (hDC, "r %+1.1f", fs->rollrate);
	SetTextPos (hDC, cw, H*1/3 + ch*8);
	TextXY (hDC, "y %+1.1f", fs->yawrate);
}

void SpaceShipOneMFD::DrawApogeeData (HDC hDC) {
	double apogee;
	pV->GetApDist(apogee);
	apogee -= 6371010;
	apogee *= 0.0032808399;
	SetMFDColor (hDC, WHITE);
	SetTextPos (hDC, W*5/6-cw*3/2, H*2/3 + ch*3/2);
	TextXY (hDC, "%3.0f", apogee);
}

void SpaceShipOneMFD::DrawPropellantData(HDC hDC) {

	double burnTime, maxThrust, ISP, fuelMass, maxFuelMass;
	PROPELLANT_HANDLE ph;
	THRUSTER_HANDLE th;

	if (pV->GetPropellantCount() < 1 || pV->GetThrusterCount() < 1) 
		return;

	ph			= pV->GetPropellantHandleByIndex(0);
	th			= pV->GetThrusterHandleByIndex(0);
	fuelMass	= pV->GetPropellantMass(ph);
	maxThrust	= pV->GetThrusterMax0(th);
	ISP			= pV->GetThrusterIsp0(th);

	if (maxThrust > 0)
		burnTime = ISP * fuelMass / maxThrust;
	else
		burnTime = 0.0;

	SetMFDColor (hDC, WHITE);
	SetTextPos (hDC, cw, H*5/6 - ch/2);
	TextXY (hDC, "BurnT: %.0fs", burnTime);
	
	if (pV->GetPropellantCount() < 2 || pV->GetThrusterCount() < 2) 
		return;

	ph			= pV->GetPropellantHandleByIndex(1);
	fuelMass	= pV->GetPropellantMass(ph);
	maxFuelMass = pV->GetPropellantMaxMass(ph);

	SetTextPos (hDC, W-cw*10, H*5/6 - ch/2);
	if (maxFuelMass > 0)
		TextXY (hDC, "RCS: %2.0f%%", fuelMass*100/maxFuelMass);

}

void SpaceShipOneMFD::DrawScale (HDC hDC, int x, int y, int r, HPEN pen, HBRUSH brush) {
	SelectObject (hDC, pen);
	SelectObject (hDC, brush);
	int dotSize = ch/4;
	int dotX, dotY;
	for (double i=0; i < 360; i+=36) {
		dotX = (int)(RoundDouble (x + sin(i*RAD)*r, 0));
		dotY = (int)(RoundDouble (y - cos(i*RAD)*r, 0));
		Ellipse (hDC, dotX-dotSize, dotY-dotSize, dotX+dotSize, dotY+dotSize);
	}
}

void SpaceShipOneMFD::DrawPointer (HDC hDC, int x, int y, int r, float angle, HPEN pen, HBRUSH brush) {
	SelectObject (hDC, pen);
	SelectObject (hDC, brush);
	POINT pts[3];

	pts[0].x = (int)(RoundValue ((float)(x + sin((angle-5)*RAD)*(r*3/4))));
	pts[0].y = (int)(RoundValue ((float)(y - cos((angle-5)*RAD)*(r*3/4))));

	pts[1].x = (int)(RoundValue ((float)(x + sin((angle+5)*RAD)*(r*3/4))));
	pts[1].y = (int)(RoundValue ((float)(y - cos((angle+5)*RAD)*(r*3/4))));

	pts[2].x = (int)(RoundValue ((float)(x + sin(angle*RAD)*r)));
	pts[2].y = (int)(RoundValue ((float)(y - cos(angle*RAD)*r)));

	Polygon (hDC, pts, 3);

}


void SpaceShipOneMFD::DrawArc (HDC hDC, int x, int y, int r, float angle, HPEN penPos, HPEN penNeg) {
	if (angle > 180) angle = 180;
	else if (angle < -180) angle = -180;
	if (angle > 0) {
		SelectObject (hDC, penPos);
		Arc(hDC, x-r, y-r, x+r, y+r, (int)(x-cos(angle*RAD)*r), (int)(y-sin(angle*RAD)*r), x-r, y);
	} else {
		SelectObject (hDC, penNeg);
		Arc(hDC, x-r, y-r, x+r, y+r, x-r, y, (int)(x-cos(angle*RAD)*r), (int)(y-sin(angle*RAD)*r));
	}
}

void SpaceShipOneMFD::DrawRates (HDC hDC) {
	SelectObject (hDC, g_GDI.pScale);
	SelectObject (hDC, GetStockObject(NULL_BRUSH));
//	MoveToEx (hDC, W/2 + W/10, H/3 - H/10, 0);
//	LineTo (hDC, W/2 - W/10, H/3 - H/10);
	int x1, x2, y1, y2;
	float sqrt2 = sqrt (2.0f);
	x1 = RoundValue (W/2 - W/10*sqrt2);
	y1 = RoundValue (H/3 - W/10*sqrt2) - 1;
	x2 = RoundValue (W/2 + W/10*sqrt2);
	y2 = RoundValue (H/3 + W/10*sqrt2) - 1;
	Arc (hDC, x1, y1, x2, y2, W/2+W/10, H/3-H/10, W/2-W/10, H/3-H/10);
	MoveToEx (hDC, W/2 - W/10, H/3 - H/10, 0);
	LineTo (hDC, W/2 - W/10, H/3 + H/10);
	LineTo (hDC, W/2 + W/10, H/3 + H/10);
	Ellipse (hDC, W/2 - W/10, H/3 - H/10, W/2 + W/10, H/3 + H/10);

	SelectObject (hDC, g_GDI.pPointer);
	SelectObject (hDC, g_GDI.bPointer);

	POINT pts[3];

	float rate = (float)fs->pitchrate;
	CheckRange (rate, -10.0f, 10.0f);
	rate *= H/5 / 20;
	pts[0].x = W/2 - W/10 - 1;
	pts[1].x = pts[0].x - cw - 1;
	pts[2].x = pts[0].x - cw - 1;

	pts[0].y = RoundValue((float)(H/3 - rate));
	pts[1].y = pts[0].y - cw/4;
	pts[2].y = pts[0].y + cw/4;
	Polygon (hDC, pts, 3);

	rate = (float)fs->yawrate;
	CheckRange (rate, -20.0f, 20.0f);
	rate *= W/5 / 10;
	pts[0].x = RoundValue((float)(W/2 + rate));
	pts[1].x = pts[0].x - cw/4;
	pts[2].x = pts[0].x + cw/4;

	pts[0].y = H/3 + H/10 + 1;
	pts[1].y = H/3 + H/10 + cw + 1;
	pts[2].y = H/3 + H/10 + cw + 1;
	Polygon (hDC, pts, 3);

	rate = (float)fs->rollrate;
	CheckRange (rate, -30.0f, 30.0f);
	rate *= 1.5f;

	float r = W/10*sqrt2;
	float r2 = ((W/10)+cw*2/3)*sqrt2;
	float spread = 2.5f;

	pts[0].x = RoundValue(W/2 + sin(rate*RADf)*r);
	pts[1].x = RoundValue(W/2 + sin((rate+spread)*RADf)*r2);
	pts[2].x = RoundValue(W/2 + sin((rate-spread)*RADf)*r2);

	pts[0].y = RoundValue(H/3 - cos(rate*RADf)*r - 2);
	pts[1].y = RoundValue(H/3 - cos((rate+spread)*RADf)*r2 - 2);
	pts[2].y = RoundValue(H/3 - cos((rate-spread)*RADf)*r2 - 2);

	Polygon (hDC, pts, 3);


}